from __future__ import annotations

import os
from dataclasses import dataclass
from enum import Enum
from typing import Any

from .config_path import ConfPath


class OverrideActionEnum(Enum):
    Default = "?="
    Assign = "="
    Append = "+="
    AppendPath = "+=(path)"
    Prepend = "=+"
    PrependPath = "=+(path)"
    Remove = "-="
    Clear = "=!"

    def override_string(self, old: str, new: str) -> str:
        return {
            OverrideActionEnum.Default: lambda: new,
            OverrideActionEnum.Assign: lambda: new,
            OverrideActionEnum.Append: lambda: old + new,
            OverrideActionEnum.AppendPath: lambda: (
                old + os.pathsep + new if old and new else old + new
            ),
            OverrideActionEnum.Prepend: lambda: new + old,
            OverrideActionEnum.PrependPath: lambda: (
                new + os.pathsep + old if old and new else old + new
            ),
            OverrideActionEnum.Remove: lambda: old.replace(new, ""),
        }[self]()


@dataclass
class OverrideAction:
    action: OverrideActionEnum
    values: Any


class ValueReference:
    def __init__(
        self,
        value_path: ConfPath,
        values: dict | OverrideAction | Any,
        action: OverrideActionEnum = OverrideActionEnum.Default,
    ) -> None:
        self.value_path = value_path
        self.action = action
        self.values: dict | Any
        if isinstance(values, OverrideAction):
            self.action = values.action
            self.values = values.values
        else:
            self.values = values

    def is_value_set(self, path: str | ConfPath):
        values = self.values
        if values is None:
            return False
        if isinstance(path, str):
            return path in values
        while path:
            name, path = path.split_front()
            if values is None or name not in values:
                return False
            values = values[name]
        return True

    def get_value(self, path: str | ConfPath):
        values = self.values
        if isinstance(path, str):
            if values is None:
                raise KeyError(path)
            return values[path]
        while path:
            name, path = path.split_front()
            if values is None:
                raise KeyError(name)
            values = values[name]
        return values

    def set_value(self, path: str | ConfPath, val: Any):
        values = self.values
        if isinstance(path, str):
            if values is None:
                return False
            values[path] = val
            return True
        while True:
            if values is None:
                return False
            name, path = path.split_front()
            if not path:
                values[name] = val
                return True
            if name not in values:
                return False
            values = values[name]

    def clear_value(self, path: str | ConfPath):
        values = self.values
        if isinstance(path, str):
            if values is None:
                return
            values.pop(path, None)
            return
        while True:
            name, path = path.split_front()
            if values is None:
                return
            if not path:
                values.pop(name, None)
                return
            if name not in values:
                return
            values = values[name]

    def set_value_default(self, path: str | ConfPath, val: Any):
        if self.values is None:
            return None
        if isinstance(path, str):
            self.values.setdefault(path, val)
            return True
        values = self.values
        while True:
            if values is None:
                return False
            name, path = path.split_front()
            if not path:
                values.setdefault(name, val)
                return True
            if name not in values:
                return False
            values = values[name]

    def sub_ref(self, name: str | ConfPath) -> ValueReference:
        if isinstance(name, ConfPath) and len(name.pth) == 1:
            name = name.pth[0]
        if isinstance(name, str):
            if not self.is_value_set(name):
                raise KeyError(name)
            return ValueReference(
                value_path=self.value_path.join(name),
                values=self.values[name],
            )
        if not name:
            return self
        name, rem = name.split_front()
        return self.sub_ref(name).sub_ref(rem)

    def __repr__(self) -> str:
        return f"<ValueReference to: '{self.value_path}'>"



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import List, cast

from lark import Lark, Token, Transformer, v_args

grammar = Path(__file__).with_suffix(".lark").read_text()


@dataclass
class CLIOption:
    action: str
    key: tuple[str, ...]
    value: bool | str | list | dict | float | None


class TreeToCLIOption(Transformer):
    @v_args(inline=True)
    def escaped_string(self, s: Token):
        return s[1:-1].encode("raw_unicode_escape").decode("unicode_escape")

    @v_args(inline=True)
    def unquoted_string(self, s: Token):
        return s.value

    @v_args(inline=True)
    def full(self, k, a, v=None):
        return CLIOption(a.value, k, v)

    lines = list
    keys = tuple

    array = list
    pair = tuple
    object = dict

    NUMBER = int
    TRUE = lambda self, _: True
    FALSE = lambda self, _: False


cli_parser = Lark(grammar, start="option", parser="lalr", transformer=TreeToCLIOption())
file_parser = Lark(grammar, start="lines", parser="lalr", transformer=TreeToCLIOption())


def parse_cli(s: str) -> CLIOption:
    return cast(CLIOption, cli_parser.parse(s))


def parse_file(s: str) -> list[CLIOption]:
    return cast(List[CLIOption], file_parser.parse(s))



from __future__ import annotations

from copy import deepcopy

from ...common import ConfigError
from .config_option import ConfigOption
from .string import StringOption
from .value_reference import OverrideActionEnum, ValueReference


class DictOfStrConfigOption(ConfigOption):
    def get_typename(self, md: bool = False) -> str:
        return "dict"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if new_value.values is None:
            return None
        if old_value.values is None:
            old_value.values = {}
        r = deepcopy(old_value.values)
        for k, v in new_value.values.items():
            if k not in r:
                r[k] = deepcopy(v)
            else:
                r[k] = r[k].override(v)
        return r

    def _verify_str(self, values: ValueReference):
        if not isinstance(values.values, str):
            msg = f"Type of values in {values.value_path} should be str, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)

    def verify(self, values: ValueReference):
        if values.values is None:
            return None
        if values.action not in (OverrideActionEnum.Append, OverrideActionEnum.Default):
            msg = f"Option {values.value_path} of type {self.get_typename()} "
            msg += f"does not support operation {values.action.value}"
            raise ConfigError(msg)

        if not isinstance(values.values, dict):
            msg = f"Type of {values.value_path} should be {dict}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        elif not all(isinstance(el, str) for el in values.values):
            msg = f"Type of keys in {values.value_path} should be {str}"
            raise ConfigError(msg)
        for k in values.values:
            self._verify_str(values.sub_ref(k))

        valdict: dict[str, StringOption] = {}
        for k in values.values:
            valdict[k] = StringOption.from_values(values.sub_ref(k))
        return valdict

    def finalize(self, values: ValueReference) -> dict[str, str] | None:
        if values.values is None:
            return None
        options: dict[str, str] = {}
        for k, v in values.values.items():
            if v is not None:
                assert isinstance(v, StringOption)
                str_v = v.finalize()
                if str_v is not None:
                    options[k] = str_v
        return options



from __future__ import annotations

from copy import copy

from ...common import ConfigError
from .config_reference import ConfigReference
from .value_reference import OverrideActionEnum, ValueReference


class ConfigVerifier:
    def __init__(
        self, root: ConfigReference, ref: ConfigReference, values: ValueReference
    ) -> None:
        self.root = root
        self.ref = ref
        self.values = values

    def verify(self):
        # If this option should be cleared, it shouldn't have a value
        if (
            self.values.action == OverrideActionEnum.Clear
            and self.values.values is not None
        ):
            msg = f'Operation "clear" ({self.values.action.value}) '
            msg += f"cannot have a value in {self.values.value_path}"
            raise ConfigError(msg)
        # Verify our own option
        verified_values = ValueReference(
            self.values.value_path,
            self.ref.config.verify(copy(self.values)),
            self.values.action,
        )
        # Verify our sub-options
        for ref, sub_val in self.ref.iter_set_sub_options(verified_values):
            verified = ConfigVerifier(
                root=self.root,
                ref=ref.resolve_inheritance(self.root),
                values=sub_val,
            ).verify()
            rel = sub_val.value_path.relative_to(verified_values.value_path)
            verified_values.set_value(rel, verified)
        return verified_values.values



from __future__ import annotations

from ...common import ConfigError
from .config_option import ConfigOption
from .value_reference import OverrideActionEnum, ValueReference


class IntConfigOption(ConfigOption):
    def get_typename(self, md: bool = False):
        return "int"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        return new_value.values

    def verify(self, values: ValueReference):
        if values.values is None:
            return None
        if values.action not in (OverrideActionEnum.Assign, OverrideActionEnum.Default):
            msg = f"Option {values.value_path} of type {self.get_typename()} "
            msg += f"does not support operation {values.action.value}"
            raise ConfigError(msg)
        elif not isinstance(values.values, int):
            msg = f"Type of {values.value_path} should be {int}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        return values.values



from __future__ import annotations

from copy import copy, deepcopy
from difflib import get_close_matches
from typing import Any, Iterable

from ...common import ConfigError
from .config_path import ConfPath
from .default import DefaultValue, NoDefaultValue
from .value_reference import OverrideActionEnum, ValueReference

_MAGIC_CLEAR_VALUE = None


class ConfigOption:
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        sub_options: dict[str, ConfigOption] | None = None,
    ) -> None:
        if default is None:
            default = NoDefaultValue()
        if isinstance(inherit_from, str):
            inherit_from = ConfPath.from_string(inherit_from)
        if sub_options is None:
            sub_options = {}
        self.name = name
        self.description = description
        self.example = example
        self.sub_options: dict[str, ConfigOption] = sub_options
        self.inherits = inherit_from
        self.default: DefaultValue = default
        self.create_if_inheritance_target_exists = create_if_inheritance_target_exists

    def insert(self, opt: ConfigOption):
        assert opt.name not in self.sub_options
        self.sub_options[opt.name] = opt
        return self.sub_options[opt.name]

    def insert_multiple(self, opts: Iterable[ConfigOption]):
        for opt in opts:
            self.insert(opt)
        return self

    def get_typename(self, md: bool = False) -> str | None:
        return None

    def iter_sub_options(
        self, values: ValueReference
    ) -> Iterable[tuple[str, ConfPath]]:
        """Return the names of the sub-options of this config option, along with
        the corresponding value path relative to the given value pth."""
        for name in self.sub_options:
            yield name, values.value_path.join(name)

    def iter_set_sub_options(
        self, values: ValueReference
    ) -> Iterable[tuple[str, ValueReference]]:
        """Return the names of the sub-options of this config option, along with
        the corresponding value reference relative to the given value if set."""
        for name in self.sub_options:
            if values.is_value_set(name):
                yield name, values.sub_ref(name)

    def verify(self, values: ValueReference) -> Any:
        if values.action == OverrideActionEnum.Clear:
            return _MAGIC_CLEAR_VALUE
        if values.action not in (OverrideActionEnum.Assign, OverrideActionEnum.Default):
            msg = f"Option {values.value_path} does not support "
            msg += f"operation {values.action.value}"
            raise ConfigError(msg)
        if not isinstance(values.values, dict):
            msg = f"Type of {values.value_path} should be 'dict', "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        unknown_keys = set(values.values) - set(self.sub_options)
        msg = ""
        for k in sorted(unknown_keys):
            suggested = get_close_matches(k, self.sub_options, 3)
            msg = f"Unknown option '{k}' in {values.value_path}. "
            if suggested:
                msg += f"Did you mean: {', '.join(suggested)}\n"
        if msg:
            raise ConfigError(msg[:-1])
        return values.values

    def finalize(self, values: ValueReference) -> Any:
        if self.sub_options:
            for k in list(values.values):
                if values.values[k] == _MAGIC_CLEAR_VALUE:
                    del values.values[k]
        return values.values

    def override(self, old_value: ValueReference, new_value: ValueReference) -> Any:
        if new_value.values == _MAGIC_CLEAR_VALUE:
            return _MAGIC_CLEAR_VALUE
        return copy(old_value.values)


class MultiConfigOption(ConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            example=example,
            default=default,
            inherit_from=inherit_from,
            create_if_inheritance_target_exists=create_if_inheritance_target_exists,
        )

    default_index = "*"

    def iter_sub_options(
        self, values: ValueReference
    ) -> Iterable[tuple[str, ConfPath]]:
        """Return the names of the sub-options of this config option, along with
        the corresponding value path relative to the given value pth."""
        if values.values is None:
            return ()
        for k in values.values:
            val = values.sub_ref(k)
            for name in self.sub_options:
                yield name, val.value_path.join(name)

    def iter_set_sub_options(
        self, values: ValueReference
    ) -> Iterable[tuple[str, ValueReference]]:
        """Return the names of the sub-options of this config option, along with
        the corresponding value reference relative to the given value if set."""
        if values.values is None:
            return ()
        for k in values.values:
            val = values.sub_ref(k)
            for name in self.sub_options:
                if val.is_value_set(name):
                    yield name, val.sub_ref(name)

    def _verify(self, values: ValueReference):
        if values.values is None:
            return
        unknown_keys = set(values.values) - set(self.sub_options)
        msg = ""
        for k in sorted(unknown_keys):
            suggested = get_close_matches(k, self.sub_options, 3)
            msg = f"Unknown option '{k}' in {values.value_path}. "
            if suggested:
                msg += f"Did you mean: {', '.join(suggested)}\n"
        if msg:
            raise ConfigError(msg[:-1])

    def _values_to_index_dict(self, values: ValueReference):
        vals: dict[str, dict | None] = {}
        for k in values.values:
            try:
                int(k)
            except ValueError:
                v0 = vals.setdefault(self.default_index, {})
                assert v0 is not None
                v0[k] = values.values[k]  # Copy any action as well
                continue
            vr = values.sub_ref(k)
            if vr.action == OverrideActionEnum.Clear:
                vals[k] = _MAGIC_CLEAR_VALUE
            elif vr.action not in (
                OverrideActionEnum.Assign,
                OverrideActionEnum.Default,
            ):
                msg = f"Option {vr.value_path} does not support "
                msg += f"operation {vr.action.value}"
                raise ConfigError(msg)
            else:
                vals[k] = vr.values  # Do not copy action
        return vals

    def verify(self, values: ValueReference) -> Any:
        if values.action == OverrideActionEnum.Clear:
            return _MAGIC_CLEAR_VALUE
        if values.action not in (OverrideActionEnum.Assign, OverrideActionEnum.Default):
            msg = f"Option {values.value_path} does not support "
            msg += f"operation {values.action.value}"
            raise ConfigError(msg)
        if not isinstance(values.values, dict):
            msg = f"Type of {values.value_path} should be 'dict', "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        if not all(isinstance(v, str) for v in values.values):
            msg = f"Type of keys in {values.value_path} should be 'str'"
            raise ConfigError(msg)
        vals = self._values_to_index_dict(values)
        for k, v in vals.items():
            self._verify(ValueReference(values.value_path.join(k), v))
        return vals

    def finalize(self, values: ValueReference) -> Any:
        from .config_reference import ConfigReference
        from .override import ConfigOverrider

        if values.is_value_set(self.default_index):
            # Our own option as a "single" ConfigOption
            opt = ConfigOption(
                name=self.name,
                description=self.description,
                example=self.example,
                default=self.default,
                inherit_from=self.inherits,
                create_if_inheritance_target_exists=self.create_if_inheritance_target_exists,
                sub_options=self.sub_options,
            )
            # Override the non-default items by the default item.
            ref = ConfigReference(ConfPath((self.name,)), opt)
            for k in list(values.values):
                if k == self.default_index:
                    continue
                if values.values[k] == _MAGIC_CLEAR_VALUE:
                    del values.values[k]
                    continue
                super_value = deepcopy(values.sub_ref(self.default_index))
                new_values = values.sub_ref(k)
                for name in self.sub_options:
                    if new_values.is_value_set(name):
                        if super_value.is_value_set(name):
                            new_value = ConfigOverrider(
                                root=None,
                                ref=ref.sub_ref(name),
                                values=super_value.sub_ref(name),
                                new_values=new_values.sub_ref(name),
                            ).override()
                        else:
                            new_value = new_values.sub_ref(name).values
                        super_value.set_value(name, new_value)
                values.set_value(k, super_value.values)
            if len(values.values) == 1:
                values.values["0"] = values.values.pop(self.default_index)
            else:
                values.clear_value(self.default_index)

        return values.values

    def override(self, old_value: ValueReference, new_value: ValueReference) -> Any:
        if new_value.values == _MAGIC_CLEAR_VALUE:
            return _MAGIC_CLEAR_VALUE
        # TODO: change implementation such that shallow copy is sufficient
        r = deepcopy(old_value.values) or {}
        for k, v in new_value.values.items():
            if v == _MAGIC_CLEAR_VALUE:
                r[k] = _MAGIC_CLEAR_VALUE
            else:
                r.setdefault(k, {})
        return r


class UncheckedConfigOption(ConfigOption):
    def verify(self, values: ValueReference):
        return values.values



from __future__ import annotations

from copy import copy

from .config_reference import ConfigReference
from .value_reference import ValueReference


class ConfigFinalizer:
    def __init__(
        self, root: ConfigReference, ref: ConfigReference, values: ValueReference
    ) -> None:
        self.root = root
        self.ref = ref
        self.values = values

    def finalize(self):
        final_values = ValueReference(
            self.values.value_path,
            self.ref.config.finalize(copy(self.values)),
        )
        for ref, val_ref in self.ref.iter_set_sub_options(final_values):
            final_val = ConfigFinalizer(
                root=self.root,
                ref=ref.resolve_inheritance(self.root),
                values=val_ref,
            ).finalize()
            rel = val_ref.value_path.relative_to(final_values.value_path)
            if final_val is None:
                final_values.clear_value(rel)
            else:
                final_values.set_value(rel, final_val)
        return final_values.values



from __future__ import annotations

import logging
from copy import copy, deepcopy
from dataclasses import dataclass
from typing import List, Union, cast

from ...common import ConfigError
from .config_option import ConfigOption
from .config_path import ConfPath
from .value_reference import OverrideActionEnum, ValueReference

logger = logging.getLogger(__name__)


@dataclass
class CMakeOption:
    clear: bool = False
    value: list[str | bool] | None = None
    append: list[str | bool] | None = None
    prepend: list[str | bool] | None = None
    remove: list[str | bool] | None = None
    datatype: str | None = None
    strict: bool = False

    @classmethod
    def create(cls, x, datatype=None):
        assert datatype is None or isinstance(datatype, str)
        if isinstance(x, list):
            assert all(isinstance(e, (bool, str)) for e in x)
            return cls(value=x, datatype=datatype)
        assert isinstance(x, (bool, str))
        return cls(value=[x], datatype=datatype)

    @classmethod
    def _from_dict(cls, x: dict[str, str | bool | list[str | bool]], pth: ConfPath):
        opt = cls()
        if "=" in x:
            assert "value" not in x
            x["value"] = x.pop("=")
        if "+" in x:
            assert "append" not in x
            x["append"] = x.pop("+")
        if "-" in x:
            assert "remove" not in x
            x["remove"] = x.pop("-")
        all_bool = True
        nonempty = False
        # Preserve lists and convert all values to singletons
        for k in ("value", "append", "remove", "prepend"):
            if k not in x:
                continue
            v = x.pop(k)
            v = v if isinstance(v, list) else [v]
            if not all(isinstance(b, bool) for b in v):
                all_bool = False
            if len(v) > 0:
                nonempty = True
            setattr(opt, k, v)
        # Determine the data type
        datatype = x.pop("type", None)
        if datatype is None and all_bool and nonempty:
            datatype = "BOOL"
        if datatype is not None and not isinstance(datatype, str):
            msg = f'Error in {pth}: Type of "type" should be str, '
            msg += f"not {type(datatype)}"
            raise ConfigError(msg)
        opt.datatype = datatype
        # Check if the data type should be checked in strict mode or not
        strict = x.pop("strict", True)
        if not isinstance(strict, bool):
            msg = f'Error in {pth}: Type of "strict" should be bool, '
            msg += f"not {type(strict)}"
            raise ConfigError(msg)
        opt.strict = strict
        # Check if there are any unused keys left
        assert not x
        return opt

    @classmethod
    def from_values(cls, values: ValueReference):
        val = values.values
        if values.action == OverrideActionEnum.Clear:
            assert val is None
            return cls(clear=True)
        if isinstance(val, dict):
            return cls._from_dict(val, values.value_path)
        datatype = None
        if isinstance(val, bool):
            val = [val]
            datatype = "BOOL"
        elif isinstance(val, str):
            val = [val]
        elif isinstance(val, list):
            assert all(isinstance(b, (str, bool)) for b in val)
            all_bool = all(isinstance(b, bool) for b in val)
            nonempty = len(val) > 0
            if all_bool and nonempty:
                datatype = "BOOL"
        else:
            msg = f"Invalid type {type(val)}"
            raise AssertionError(msg)
        val = cast(List[Union[str, bool]], val)
        if values.action in (OverrideActionEnum.Default, OverrideActionEnum.Assign):
            return cls(value=val, datatype=datatype)
        if values.action == OverrideActionEnum.Remove:
            return cls(remove=val, datatype=datatype)
        if values.action == OverrideActionEnum.Append:
            return cls(append=val, datatype=datatype)
        if values.action == OverrideActionEnum.Prepend:
            return cls(prepend=val, datatype=datatype)
        msg = f"Invalid action {values.action}"
        raise AssertionError(msg)

    @staticmethod
    def _combine_types(
        a: CMakeOption,
        b: CMakeOption,
        strict: bool,
        a_path: ConfPath,
        b_path: ConfPath,
    ) -> str | None:
        def report(msg):
            if not strict:
                logger.warning(msg)
                return b.datatype
            msg = f'{msg} (use "strict" = false to ignore)'
            raise ConfigError(msg)

        if b.datatype is None:
            return a.datatype
        if a.datatype is None:
            return b.datatype
        if a.datatype == b.datatype:
            return b.datatype
        msg = "Incompatible types when overriding or inheriting CMake settings"
        # Cannot override these by other types
        if a.datatype in ("BOOL", "STRING", "FILEPATH"):
            return report(f"{msg}: {a.datatype} ({a_path}) and {b.datatype} ({b_path})")
        # PATH can be made more specific to FILEPATH
        if a.datatype == "PATH" and b.datatype != "FILEPATH":
            return report(f"{msg}: {a.datatype} ({a_path}) and {b.datatype} ({b_path})")
        return b.datatype

    def override(self, new: CMakeOption):
        # Clearing always propagates
        if new.clear:
            return copy(new)
        old = copy(self)
        # If we're overriding with a value, the old value is not used
        if new.value is not None:
            old.value = new.value
            old.append = new.append
            old.prepend = new.prepend
            old.remove = []
            return old
        # Otherwise, we're simply appending to or removing from the old value,
        # and these changes are propagated down to the old value.
        if new.remove is not None:
            old.remove = (old.remove or []) + new.remove
            if old.value:
                old.value = [e for e in old.value if e not in new.remove]
            if old.prepend:
                old.prepend = [e for e in old.prepend if e not in new.remove]
            if old.append:
                old.append = [e for e in old.append if e not in new.remove]
        if new.append is not None:
            old.append = (old.append or []) + new.append
        if new.prepend is not None:
            old.prepend = new.prepend + (old.prepend or [])
        return old

    def finalize(self) -> list[str | bool] | None:
        empty = True
        final = []
        if self.value is not None:
            empty = False
            final = self.value
        if self.append is not None:
            empty = False
            final = final + self.append
        if self.prepend is not None:
            empty = False
            final = self.prepend + final
        return None if (empty and self.clear) else final


# Entries are of the following form
# {
#    "VAR_A": "value",
#    "VAR_B": ["123", "456"],
#    "VAR_C": false,
#    "VAR_D": {
#       "type": "STRING",
#       "value": ["789", "xyz"],
#    },
#    "VAR_E": {
#       "type": "FILEPATH",
#       "value": "/usr/bin/bash",
#    },
#    "VAR_F": {
#       "type": "FILEPATH",
#       "append": ["/usr/bin/bash"],
#    }
# }


class CMakeOptConfigOption(ConfigOption):
    """
    Dictionary of CMake options, with optional data type, ability to merge
    or remove options, etc.
    """

    valid_keys = frozenset(
        ("+", "-", "=", "append", "remove", "value", "prepend", "type", "strict")
    )

    def get_typename(self, md: bool = False) -> str:
        return "dict (CMake)"

    @staticmethod
    def _check_type(a, a_path):
        valid_types = (None, "STRING", "BOOL", "PATH", "FILEPATH")
        if a not in valid_types:
            msg = f"Unknown type '{a}' in {a_path}"
            raise ConfigError(msg)

    def _combine(self, a: CMakeOption, b: CMakeOption, a_path, b_path):
        r = a.override(b)
        r.datatype = CMakeOption._combine_types(a, b, b.strict, a_path, b_path)
        return r

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if new_value.values is None:
            return None
        if old_value.values is None:
            old_value.values = {}
        r = deepcopy(old_value.values)
        for k, v in new_value.values.items():
            if k not in r:
                r[k] = deepcopy(v)
            else:
                r[k] = self._combine(
                    a=r[k],
                    b=v,
                    a_path=old_value.value_path.join(k),
                    b_path=new_value.value_path.join(k),
                )
        return r

    def _verify_cmake_list(self, values: ValueReference):
        if isinstance(values.values, list):
            for v in values.values:
                if not isinstance(v, str) and not isinstance(v, bool):
                    msg = f"Type of values in list {values.value_path} should "
                    msg += f"be str or bool, not {type(v)}"
                    raise ConfigError(msg)
        elif not isinstance(values.values, str) and not isinstance(values.values, bool):
            msg = f"Type of values in {values.value_path} should be str or bool or a "
            msg += f"list thereof, not {type(values.values)}"
            raise ConfigError(msg)

    def _verify_cmake_list_or_dict(self, values: ValueReference):
        if isinstance(values.values, dict):
            for k in values.values:
                if not isinstance(k, str):
                    msg = f"Type of keys in {values.value_path} should be str, "
                    msg += f"not {type(k)}"
                    raise ConfigError(msg)
                self._verify_cmake_list(values.sub_ref(k))
        else:
            self._verify_cmake_list(values)

    def _check_dict_keys(self, d, pth: ConfPath):
        if not isinstance(d, dict):
            return
        invalid_keys = set(d) - self.valid_keys
        if invalid_keys:
            msg = f"Invalid keys in {pth}: {list(invalid_keys)}"
            raise ConfigError(msg)
        if "value" in d and "=" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "value" and "="'
            raise ConfigError(msg)
        if "append" in d and "+" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "append" and "+"'
            raise ConfigError(msg)
        if "remove" in d and "-" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "remove" and "-"'
            raise ConfigError(msg)
        if "value" in d or "=" in d:
            invalid_keys = {"+", "append", "-", "remove", "prepend"} & set(d)
            if invalid_keys:
                msg = f"Invalid keys in {pth}: "
                msg += 'Cannot combine "value" or "=" with the '
                msg += f"following keys: {list(invalid_keys)}"
                raise ConfigError(msg)
        if "type" in d:
            self._check_type(d["type"], pth)

    def verify(self, values: ValueReference):
        """Checks the data types and keys of the values, and then converts
        them to a dictionary of CMakeOptions."""
        if values.values is None:
            return None
        if values.action not in (OverrideActionEnum.Append, OverrideActionEnum.Default):
            msg = f"Option {values.value_path} of type {self.get_typename()} "
            msg += f"does not support operation {values.action.value}"
            raise ConfigError(msg)
        if not isinstance(values.values, dict):
            msg = f"Type of {values.value_path} should be {dict}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        elif not all(isinstance(k, str) for k in values.values):
            msg = f"Type of keys in {values.value_path} should be {str}"
            raise ConfigError(msg)
        for k in values.values:
            self._verify_cmake_list_or_dict(values.sub_ref(k))

        valdict: dict[str, CMakeOption] = {}
        for k in values.values:
            pth = values.value_path.join(k)
            self._check_dict_keys(values.values[k], pth)
            valdict[k] = CMakeOption.from_values(values.sub_ref(k))
        return valdict

    def finalize(self, values: ValueReference) -> dict[str, str] | None:
        """Converts the internal dict of CMakeOptions back into a dict of string
        that can be passed as CMake command-line arguments"""
        if values.values is None:
            return None

        def convert(x):
            assert isinstance(x, CMakeOption)
            final = x.finalize()
            if final is None:
                return None
            on_off = {True: "On", False: "Off"}
            escape = lambda x: x.replace(";", "\\;")
            to_str = lambda b: (escape(b) if isinstance(b, str) else on_off[b])
            return ";".join(map(to_str, final))

        def convert_key(k, x):
            assert isinstance(x, CMakeOption)
            return f"{k}:{x.datatype}" if x.datatype else k

        options: dict[str, str] = {}
        for k, v in values.values.items():
            if v is not None:
                str_v = convert(v)
                if str_v is not None:
                    options[convert_key(k, v)] = str_v
        return options



from __future__ import annotations

from typing import Any

from .config_path import ConfPath
from .config_reference import ConfigReference
from .value_reference import ValueReference


class ConfigOverrider:
    def __init__(
        self,
        root: ConfigReference | None,
        ref: ConfigReference,
        values: ValueReference,
        new_values: ValueReference,
    ) -> None:
        self.ref = ref
        self.root = root
        self.values = values
        self.new_values = new_values

    def override(self):
        # Override our own value
        overridden_values = ValueReference(
            self.values.value_path,
            self.ref.config.override(self.values, self.new_values),
        )
        # If we have sub-options, override those
        for ref, new_val in self.ref.iter_set_sub_options(self.new_values):
            if self.root is not None:
                ref = ref.resolve_inheritance(self.root)  # noqa: PLW2901
            rel = new_val.value_path.relative_to(self.new_values.value_path)
            # Create default parent options if necessary (for MultiConfigOption)
            if not self._create_default_parents(overridden_values, ref, rel):
                continue
            # Replace the old value by the override
            old_val = overridden_values.sub_ref(rel)
            overridden_values.set_value(
                rel,
                ConfigOverrider(
                    root=self.root,
                    ref=ref,
                    values=old_val,
                    new_values=new_val,
                ).override(),
            )
        return overridden_values.values

    def _create_default_parents(
        self, overridden_values: ValueReference, ref: ConfigReference, rel: ConfPath
    ):
        fst, rem = rel.split_front()
        relp = ConfPath((fst,))
        for p in rem.pth:
            if not overridden_values.set_value_default(relp, {}):
                return False
            relp = relp.join(p)
        default: Any = {} if ref.sub_options else None
        return overridden_values.set_value_default(relp, default)



"""
distlib.wheel doesn't always return the correct tags, and packaging.tags
returns all tags supported by the interpreter, not the tags that should be used
for the generated wheels. Therefore the only option here is to write our own
(kind of hacky) functions based on packaging.tags.
"""

from __future__ import annotations

import os
import platform
import re
import sys
import sysconfig
from importlib.machinery import EXTENSION_SUFFIXES
from typing import Dict, List, Mapping

from distlib.util import (  # type: ignore[import-untyped]
    get_platform as get_platform_dashes,
)
from packaging import tags

from ..common.util import archflags_to_platform_tag, platform_to_platform_tag

_INTERPRETER_SHORT_NAMES: dict[str, str] = {
    "python": "py",
    "cpython": "cp",
    "pypy": "pp",
    "ironpython": "ip",
    "jython": "jy",
}


def get_platform_tag() -> str:
    return platform_to_platform_tag(get_platform_dashes())


def get_interpreter_name() -> str:
    name = sys.implementation.name
    return _INTERPRETER_SHORT_NAMES.get(name) or name


def get_interpreter_version() -> str:
    return "".join(map(str, sys.version_info[:2]))


def get_cpython_interpreter() -> str:
    return f"cp{get_interpreter_version()}"


def get_cpython_abi() -> str:
    """
    Get the ABI string for CPython, e.g. cp37m.

    https://github.com/pypa/packaging/blob/917612f5774571a99902b5fe04d06099b9e8b667/packaging/tags.py#L135
    """
    py_version = sys.version_info[:2]
    debug = pymalloc = ""
    with_debug = sysconfig.get_config_var("Py_DEBUG")
    has_refcount = hasattr(sys, "gettotalrefcount")
    # Windows doesn't set Py_DEBUG, so checking for support of debug-compiled
    # extension modules is the best option.
    # https://github.com/pypa/pip/issues/3383#issuecomment-173267692
    has_ext = "_d.pyd" in EXTENSION_SUFFIXES
    if with_debug or (with_debug is None and (has_refcount or has_ext)):
        debug = "d"
    if py_version < (3, 8):
        with_pymalloc = sysconfig.get_config_var("WITH_PYMALLOC")
        if with_pymalloc or with_pymalloc is None:
            pymalloc = "m"
    return f"{get_cpython_interpreter()}{debug}{pymalloc}"


def get_generic_interpreter() -> str:
    return f"{get_interpreter_name()}{get_interpreter_version()}"


def get_generic_abi() -> str:
    abi = sysconfig.get_config_var("SOABI") or "none"
    return platform_to_platform_tag(abi)


def get_python_tag() -> str:
    if get_interpreter_name() == "cp":
        return get_cpython_interpreter()
    else:
        return get_generic_interpreter()


def get_abi_tag() -> str:
    if get_interpreter_name() == "cp":
        return get_cpython_abi()
    else:
        return get_generic_abi()


def _guess_arch_mac(archflags: str | None) -> str | None:
    """Returns 'arm64', 'x86_64' or 'universal2', depending on the value of the
    ARCHFLAGS environment variable if set. Otherwise returns the architecture
    obtained from platform.mac_ver(). Returns None if ARCHFLAGS was invalid."""
    if archflags:
        archs = list(re.findall(r"-arch +(\S+)", archflags))
    else:
        _, _, arch = platform.mac_ver()
        archs = [arch]
    return archflags_to_platform_tag(archs)


def _guess_version_mac(target: str | None) -> tuple[int, int]:
    """Returns the macOS version set by the MACOSX_DEPLOYMENT_TARGET environment
    variable, otherwise returns the version from platform.mac_ver()."""
    version_str = None
    if target:
        m = re.match(r"^(\d+\.\d+)", target)
        if m:
            version_str = m.group(1)
    if not version_str:
        version_str, _, _ = platform.mac_ver()
    version_parts = version_str.split(".")
    major = int(version_parts[0]) if len(version_parts) > 0 else 0
    minor = int(version_parts[1]) if len(version_parts) > 1 else 0
    return major, minor


def _guess_platform_tag_mac(env: Mapping[str, str] | None = None) -> str | None:
    """Return an appropriate platform tag based on the ARCHFLAGS and
    MACOSX_DEPLOYMENT_TARGET environment variables if set. Returns None if
    neither was set, or if the platform tag could not be determined based on
    their values."""
    if env is None:
        env = os.environ
    archflags = env.get("ARCHFLAGS")
    target = env.get("MACOSX_DEPLOYMENT_TARGET")
    if not (archflags or target):
        return None
    arch = _guess_arch_mac(archflags)
    if not arch:
        return None
    version = _guess_version_mac(target)
    if arch == "arm64":
        version = max(version, (11, 0))  # ARM64 is only supported on 11.0+
    return f"macosx_{version[0]}_{version[1]}_{arch}"


def guess_platform_tag(env=None) -> str:
    if platform.system() == "Darwin":
        plat = _guess_platform_tag_mac(env)
        if plat:
            return plat
    try:
        return next(tags.sys_tags()).platform
    except StopIteration:
        return get_platform_tag()


WheelTags = Dict[str, List[str]]


def get_native_tags(guess=False) -> WheelTags:
    """Get the PEP 425 tags for the current platform."""
    return {
        "pyver": [get_python_tag()],
        "abi": [get_abi_tag()],
        "arch": [guess_platform_tag() if guess else get_platform_tag()],
    }



from __future__ import annotations

import logging
import os
import sys
import sysconfig
from dataclasses import dataclass
from itertools import product
from pathlib import Path
from string import Template

from distlib.version import NormalizedVersion  # type: ignore[import-untyped]

from .. import __version__
from ..common import PackageInfo
from ..common.util import python_sysconfig_platform_to_cmake_platform_win
from .cmd_runner import CommandRunner

logger = logging.getLogger(__name__)


@dataclass
class CMakeSettings:
    working_dir: Path
    source_path: Path
    build_path: Path
    os: str
    find_python: bool
    find_python3: bool
    minimum_required: str
    command: Path = Path("cmake")


@dataclass
class CMakeConfigureSettings:
    environment: dict
    build_type: str | None
    options: dict[str, str]
    args: list[str]
    preset: str | None
    generator: str | None
    make_program: Path | None
    cross_compiling: bool
    toolchain_file: Path | None
    python_prefix: Path | None
    python_library: Path | None
    python_include_dir: Path | None
    python_interpreter_id: str | None


@dataclass
class CMakeBuildSettings:
    args: list[str]
    tool_args: list[str]
    presets: list[str]
    configs: list[str]


@dataclass
class CMakeInstallSettings:
    args: list[str]
    configs: list[str]
    components: list[str]
    prefix: Path | None


@dataclass
class Option:
    name: str
    value: str
    type: str = "STRING"
    description: str = ""


class CMaker:
    def __init__(
        self,
        cmake_settings: CMakeSettings,
        conf_settings: CMakeConfigureSettings,
        build_settings: CMakeBuildSettings,
        install_settings: CMakeInstallSettings,
        package_info: PackageInfo,
        runner: CommandRunner,
    ):
        self.cmake_settings = cmake_settings
        self.conf_settings = conf_settings
        self.build_settings = build_settings
        self.install_settings = install_settings
        self.package_info = package_info
        self.runner = runner
        self.environment: dict[str, str] | None = None

    def run(self, *args, **kwargs):
        return self.runner.run(*args, **kwargs)

    def prepare_environment(self):
        """Copy of the current environment with the variables defined in the
        user's configuration settings."""
        if self.environment is None:
            self.environment = os.environ.copy()
            self.environment["PY_BUILD_CMAKE_VERSION"] = str(__version__)
            self.environment["PY_BUILD_CMAKE_BINARY_DIR"] = str(
                self.cmake_settings.build_path
            )
            if self.install_settings.prefix is not None:
                self.environment["PY_BUILD_CMAKE_INSTALL_PREFIX"] = str(
                    self.install_settings.prefix
                )
            if self.conf_settings.environment:
                for k, v in self.conf_settings.environment.items():
                    templ = Template(v)
                    self.environment[k] = templ.substitute(self.environment)
        return self.environment

    def cross_compiling(self) -> bool:
        return self.conf_settings.cross_compiling

    def get_configure_options_package(self) -> list[Option]:
        """Flags specific to py-build-cmake, useful in the user's CMake scripts."""
        return [
            Option("PY_BUILD_CMAKE_VERSION", str(__version__)),
            Option("PY_BUILD_CMAKE_PACKAGE_VERSION", self.package_info.version),
            Option("PY_BUILD_CMAKE_PACKAGE_NAME", self.package_info.norm_name),
            Option("PY_BUILD_CMAKE_MODULE_NAME", self.package_info.module_name),
        ]

    def get_native_python_abi_tuple(self):
        cmake_version = self.cmake_settings.minimum_required
        has_t_flag = NormalizedVersion("3.30") <= NormalizedVersion(cmake_version)
        dmu = "dmut" if has_t_flag else "dmu"
        return ";".join("ON" if c in sys.abiflags else "OFF" for c in dmu)

    def get_native_python_implementation(self) -> str | None:
        return {
            "cpython": "CPython",
            "pypy": "PyPy",
        }.get(sys.implementation.name)

    def get_native_python_hints(self, prefix):
        """FindPython hints and artifacts for this (native) Python installation."""
        yield Option(prefix + "_ROOT_DIR", Path(sys.prefix).as_posix())
        impl = self.get_native_python_implementation()
        if impl:
            yield Option(prefix + "_FIND_IMPLEMENTATIONS", impl)
        if impl == "PyPy":
            inc = sysconfig.get_path("platinclude")
            if inc:
                yield Option(prefix + "_INCLUDE_DIR", Path(inc).as_posix())
        return
        # FIND_ABI seems to confuse CMake
        yield Option(prefix + "_FIND_ABI", self.get_native_python_abi_tuple())

    def get_cross_python_hints(self, prefix):
        """FindPython hints and artifacts to set when cross-compiling."""
        if self.conf_settings.python_prefix:
            pfx = self.conf_settings.python_prefix.as_posix()
            yield Option(prefix + "_ROOT_DIR", pfx, "PATH")
        if self.conf_settings.python_library:
            lib = self.conf_settings.python_library.as_posix()
            yield Option(prefix + "_LIBRARY", lib, "FILEPATH")
        if self.conf_settings.python_include_dir:
            inc = self.conf_settings.python_include_dir.as_posix()
            yield Option(prefix + "_INCLUDE_DIR", inc, "PATH")
        if self.conf_settings.python_interpreter_id:
            id = self.conf_settings.python_interpreter_id
            yield Option(prefix + "_INTERPRETER_ID", id, "STRING")

    def get_configure_options_python(self) -> list[Option]:
        """Flags to help CMake find the right version of Python."""

        def get_opts(prefix):
            executable = Path(sys.executable).as_posix()
            yield Option(prefix + "_EXECUTABLE", executable, "FILEPATH")
            yield Option(prefix + "_FIND_REGISTRY", "NEVER")
            yield Option(prefix + "_FIND_FRAMEWORK", "NEVER")
            yield Option(prefix + "_FIND_STRATEGY", "LOCATION")
            yield Option(prefix + "_FIND_VIRTUALENV", "FIRST")
            if not self.cross_compiling():
                yield from self.get_native_python_hints(prefix)
            else:
                yield from self.get_cross_python_hints(prefix)

        opts = []
        if self.cmake_settings.find_python:
            opts += list(get_opts("Python"))
        if self.cmake_settings.find_python3:
            opts += list(get_opts("Python3"))
        return opts

    def get_configure_options_install(self) -> list[Option]:
        prefix = self.install_settings.prefix
        if prefix:
            return [
                Option("CMAKE_INSTALL_PREFIX", prefix.as_posix(), "PATH"),
                Option("CMAKE_FIND_NO_INSTALL_PREFIX", "On", "BOOL"),
            ]
        return []

    def get_configure_options_make(self) -> list[Option]:
        """Sets CMAKE_MAKE_PROGRAM."""
        if self.conf_settings.make_program:
            opt = Option(
                "CMAKE_MAKE_PROGRAM",
                self.conf_settings.make_program.as_posix(),
                "FILEPATH",
            )
            return [opt]
        return []

    def get_configure_options_toolchain(self) -> list[str]:
        """Sets CMAKE_TOOLCHAIN_FILE."""
        return (
            ["CMAKE_TOOLCHAIN_FILE:FILEPATH=" + str(self.conf_settings.toolchain_file)]
            if self.conf_settings.toolchain_file
            else []
        )

    def get_configure_options_settings(self) -> list[str]:
        return [k + "=" + v for k, v in self.conf_settings.options.items()]

    def get_configure_options(self) -> list[str]:
        """Get the list of options (-D) passed to the CMake configure step
        through the command line."""
        return (
            self.get_configure_options_toolchain()
            + self.get_configure_options_settings()
        )

    def get_preload_options(self) -> list[Option]:
        """Get the list of options set in the CMake pre-load script (-C)."""
        return (
            self.get_configure_options_package()
            + self.get_configure_options_make()
            + self.get_configure_options_python()
            + self.get_configure_options_install()
        )

    def write_preload_options(self) -> list[str]:
        """Write the options into the CMake pre-load script and return the
        command-line flags that tell CMake to load it."""
        opts = self.get_preload_options()
        if not opts:
            return []

        def fmt_opt(o: Option):
            return (
                f'set({o.name} "{o.value}"'
                f' CACHE {o.type} "{o.description}" FORCE)\n'
            )

        preload_file = self.cmake_settings.build_path / "py-build-cmake-preload.cmake"
        version = self.cmake_settings.minimum_required
        if self.runner.verbose:
            print("Writing CMake pre-load file")
            print(f"{preload_file}")
            print("---------------------------")
            print(f"cmake_minimum_required(VERSION {version})")
            for o in opts:
                print(fmt_opt(o), end="")
            print("---------------------------\n")
        if not self.runner.dry:
            self.cmake_settings.build_path.mkdir(parents=True, exist_ok=True)
            with preload_file.open("w", encoding="utf-8") as f:
                f.write(f"cmake_minimum_required(VERSION {version})\n")
                for o in opts:
                    f.write(fmt_opt(o))
        return ["-C", str(preload_file)]

    def get_cmake_generator_platform(self) -> list[str]:
        win = self.cmake_settings.os == "windows"
        gen = self.conf_settings.generator
        vs_gen = win and (not gen or gen.lower().startswith("visual studio"))
        if vs_gen and not self.cross_compiling():
            plat = sysconfig.get_platform()
            cmake_plat = python_sysconfig_platform_to_cmake_platform_win(plat)
            if cmake_plat:
                return ["-A", cmake_plat]
            else:
                msg = "Unknown platform, CMake generator platform option (-A) will not be set"
                logger.warning(msg)
        return []

    def get_configure_command(self):
        options = self.get_configure_options()
        cmd = [str(self.cmake_settings.command)]
        cmd += ["-S", str(self.cmake_settings.source_path)]
        if self.conf_settings.preset:
            cmd += ["--preset", self.conf_settings.preset]
        if not self.conf_settings.preset or not self.build_settings.presets:
            cmd += ["-B", str(self.cmake_settings.build_path)]
        if self.conf_settings.generator:
            cmd += ["-G", self.conf_settings.generator]
        cmd += self.get_cmake_generator_platform()
        cmd += self.write_preload_options()
        cmd += [f for opt in options for f in ("-D", opt)]
        cmd += self.conf_settings.args
        return cmd

    def get_working_dir(self):
        cwd = self.cmake_settings.working_dir
        return str(cwd) if cwd is not None else None

    def configure(self):
        env = self.prepare_environment()
        cmd = self.get_configure_command()
        cwd = self.get_working_dir()
        self.run(cmd, cwd=cwd, check=True, env=env)

    def get_build_command(self, config, preset):
        cmd = [str(self.cmake_settings.command), "--build"]
        if preset is not None:
            cmd += ["--preset", preset]
        else:
            cmd += [str(self.cmake_settings.build_path)]
        if config is not None:
            cmd += ["--config", config]
        if self.build_settings.args:
            cmd += self.build_settings.args
        if self.build_settings.tool_args:
            cmd += ["--", *self.build_settings.tool_args]
        return cmd

    def get_build_commands(self):
        presets = self.build_settings.presets or [None]
        configs = self.build_settings.configs or [None]
        for preset, config in product(presets, configs):
            yield self.get_build_command(config, preset)

    def build(self):
        env = self.prepare_environment()
        cwd = self.get_working_dir()
        for cmd in self.get_build_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)

    def get_install_command(self, config):
        for component in self.install_settings.components:
            cmd = [str(self.cmake_settings.command), "--install"]
            cmd += [str(self.cmake_settings.build_path)]
            if self.install_settings.prefix:
                cmd += ["--prefix", str(self.install_settings.prefix)]
            if config is not None:
                cmd += ["--config", config]
            if component:
                cmd += ["--component", component]
            if self.install_settings.args:
                cmd += self.install_settings.args
            yield cmd

    def get_install_commands(self):
        for config in self.install_settings.configs or [None]:
            yield from self.get_install_command(config)

    def install(self):
        env = self.prepare_environment()
        cwd = self.get_working_dir()
        for cmd in self.get_install_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)



from __future__ import annotations

import os
from copy import copy
from dataclasses import dataclass

from ...common import ConfigError
from .config_option import ConfigOption
from .value_reference import OverrideActionEnum, ValueReference


@dataclass
class StringOption:
    clear: bool = False
    value: str | None = None
    append: str | None = None
    append_path: str | None = None
    prepend: str | None = None
    prepend_path: str | None = None
    remove: list[str] | None = None

    @classmethod
    def create(cls, value: str):
        return cls(value=value)

    @classmethod
    def from_values(cls, values: ValueReference):  # noqa: PLR0911
        val = values.values
        if values.action == OverrideActionEnum.Clear:
            assert val is None
            return cls(clear=True)
        # The value should be a string, unless it's a "remove" action, then
        # list[str] is also fine
        msg = f"Type of {values.value_path} should be {str}, "
        msg += f"not {type(values.values)}"
        if values.action == OverrideActionEnum.Remove:
            if isinstance(val, list):
                if not all(isinstance(v, str) for v in val):
                    raise ConfigError(msg)
            elif not isinstance(val, str):
                raise ConfigError(msg)
            val_list: list[str] = val if isinstance(val, list) else [val]
            return cls(remove=val_list)
        if not isinstance(val, str):
            raise ConfigError(msg)
        if values.action in (OverrideActionEnum.Default, OverrideActionEnum.Assign):
            return cls(value=val)
        if values.action == OverrideActionEnum.Append:
            return cls(append=val)
        if values.action == OverrideActionEnum.AppendPath:
            return cls(append_path=val)
        if values.action == OverrideActionEnum.Prepend:
            return cls(prepend=val)
        if values.action == OverrideActionEnum.PrependPath:
            return cls(prepend_path=val)
        msg = f"Invalid action {values.action}"
        raise AssertionError(msg)

    @staticmethod
    def _join_path(a, b):
        return a + os.pathsep + b if a and b else a + b

    def override(self, new: StringOption):  # noqa: PLR0912
        # Clearing always propagates
        if new.clear:
            return copy(new)
        old = copy(self)
        # If we're overriding with a value, the old value is not used
        if new.value is not None:
            old.value = new.value
            old.append = new.append
            old.append_path = new.append_path
            old.prepend = new.prepend
            old.prepend_path = new.prepend_path
            old.remove = []
            return old
        # Otherwise, we're simply appending to or removing from the old value,
        # and these changes are propagated down to the old value.
        if new.remove is not None:
            old.remove = (old.remove or []) + new.remove
            for r in new.remove:
                if old.value:
                    old.value = old.value.replace(r, "")
                if old.prepend:
                    old.prepend = old.prepend.replace(r, "")
                if old.prepend_path:
                    old.prepend_path = old.prepend_path.replace(r, "")
                if old.append:
                    old.append = old.append.replace(r, "")
                if old.append_path:
                    old.append_path = old.append_path.replace(r, "")
        if new.append is not None:
            old.append = (old.append or "") + new.append
        if new.append_path is not None:
            old.append_path = self._join_path(old.append_path or "", new.append_path)
        if new.prepend is not None:
            old.prepend = new.prepend + (old.prepend or "")
        if new.prepend_path is not None:
            old.prepend_path = self._join_path(new.prepend_path, old.prepend_path or "")
        return old

    def finalize(self) -> str | None:
        empty = True
        final = ""
        # TODO: combining append and append_path may result in weird order
        if self.value is not None:
            empty = False
            final = self.value
        if self.append is not None:
            empty = False
            final = final + self.append
        if self.append_path is not None:
            empty = False
            final = StringOption._join_path(final, self.append_path)
        if self.prepend is not None:
            empty = False
            final = self.prepend + final
        if self.prepend_path is not None:
            empty = False
            final = StringOption._join_path(self.prepend_path, final)
        return None if (empty and self.clear) else final


class StringConfigOption(ConfigOption):
    def get_typename(self, md: bool = False) -> str:
        return "string"

    def verify(self, values: ValueReference):
        return StringOption.from_values(values)

    def override(self, old_value, new_value):
        new, old = new_value.values, copy(old_value.values)
        if old is None:  # No previous value
            old = StringOption()
        assert isinstance(new, StringOption)
        assert isinstance(old, StringOption)
        return old.override(new)

    def finalize(self, values: ValueReference):
        if values.values is None:
            return None
        val = values.values
        if isinstance(val, str):
            return val
        assert isinstance(val, StringOption)
        return val.finalize()



from __future__ import annotations

from copy import copy
from dataclasses import dataclass

from ...common import ConfigError
from .config_option import ConfigOption
from .config_path import ConfPath
from .default import DefaultValue
from .string import StringOption
from .value_reference import OverrideActionEnum, ValueReference


@dataclass
class ListOption:
    clear: bool = False
    value: list[str] | None = None
    append: list[str] | None = None
    prepend: list[str] | None = None
    remove: list[str] | None = None

    @classmethod
    def from_values(cls, values: ValueReference, append_by_default: bool):
        val = values.values
        if values.action == OverrideActionEnum.Clear:
            assert val is None
            return cls(clear=True)
        if not isinstance(val, list) or not all(isinstance(v, str) for v in val):
            msg = f"Invalid type {type(val)}"
            raise AssertionError(msg)
        if values.action == OverrideActionEnum.Default:
            return cls(append=val) if append_by_default else cls(value=val)
        if values.action == OverrideActionEnum.Assign:
            return cls(value=val)
        if values.action == OverrideActionEnum.Remove:
            return cls(remove=val)
        if values.action == OverrideActionEnum.Append:
            return cls(append=val)
        if values.action == OverrideActionEnum.Prepend:
            return cls(prepend=val)
        msg = f"Invalid action {values.action}"
        raise AssertionError(msg)

    def override(self, new: ListOption):
        # Clearing always propagates
        if new.clear:
            return copy(new)
        old = copy(self)
        # If we're overriding with a value, the old value is not used
        if new.value is not None:
            old.value = new.value
            old.append = new.append
            old.prepend = new.prepend
            old.remove = []
            return old
        # Otherwise, we're simply appending to or removing from the old value,
        # and these changes are propagated down to the old value.
        if new.remove is not None:
            old.remove = (old.remove or []) + new.remove
            if old.value:
                old.value = [e for e in old.value if e not in new.remove]
            if old.prepend:
                old.prepend = [e for e in old.prepend if e not in new.remove]
            if old.append:
                old.append = [e for e in old.append if e not in new.remove]
        if new.append is not None:
            old.append = (old.append or []) + new.append
        if new.prepend is not None:
            old.prepend = new.prepend + (old.prepend or [])
        return old

    def finalize(self) -> list[str] | None:
        empty = True
        final = []
        if self.value is not None:
            empty = False
            final = self.value
        if self.append is not None:
            empty = False
            final = final + self.append
        if self.prepend is not None:
            empty = False
            final = self.prepend + final
        return None if (empty and self.clear) else final


class ListOfStrConfigOption(ConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        convert_str_to_singleton=False,
        append_by_default=False,
    ) -> None:
        super().__init__(
            name,
            description,
            example,
            default,
            inherit_from,
            create_if_inheritance_target_exists,
        )
        self.convert_str_to_singleton = convert_str_to_singleton
        self.append_by_default = append_by_default

    valid_keys = frozenset(("+", "-", "=", "append", "remove", "value", "prepend"))

    def get_typename(self, md: bool = False):
        return "list+" if self.append_by_default else "list"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        new, old = new_value.values, copy(old_value.values)
        if old is None:  # No previous value
            old = ListOption()
        assert isinstance(new, ListOption)
        assert isinstance(old, ListOption)
        return old.override(new)

    def _check_dict_keys(self, d: dict, pth: ConfPath):
        invalid_keys = set(d) - self.valid_keys
        if invalid_keys:
            msg = f"Invalid keys in {pth}: {list(invalid_keys)}"
            raise ConfigError(msg)
        if "value" in d and "=" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "value" and "="'
            raise ConfigError(msg)
        if "append" in d and "+" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "append" and "+"'
            raise ConfigError(msg)
        if "remove" in d and "-" in d:
            msg = f"Invalid keys in {pth}: "
            msg += 'Cannot combine "remove" and "-"'
            raise ConfigError(msg)
        if "value" in d or "=" in d:
            invalid_keys = {"+", "append", "-", "remove", "prepend"} & set(d)
            if invalid_keys:
                msg = f"Invalid keys in {pth}: "
                msg += 'Cannot combine "value" or "=" with the '
                msg += f"following keys: {list(invalid_keys)}"
                raise ConfigError(msg)

    def _verify_dict(self, values: ValueReference) -> ListOption:
        if values.action not in (OverrideActionEnum.Assign, OverrideActionEnum.Default):
            msg = f"Type of {values.value_path} should be {list}, "
            msg += f"not {dict}"
            raise ConfigError(msg)
        x = values.values
        self._check_dict_keys(x, values.value_path)
        opt = ListOption()
        if "=" in x:
            x["value"] = x.pop("=")
        if "+" in x:
            x["append"] = x.pop("+")
        if "-" in x:
            x["remove"] = x.pop("-")
        for k in ("value", "append", "remove", "prepend"):
            if k not in x:
                continue
            v = x.pop(k)
            pthname = f"{values.value_path}[{k}]"
            if not isinstance(v, list):
                msg = f"Type of {pthname} should be {list}, not {type(v)}"
                raise ConfigError(msg)
            if not all(isinstance(el, str) for el in v):
                msg = f"Type of elements in {pthname} should be {str}"
                raise ConfigError(msg)
            setattr(opt, k, v)
        assert not x
        return opt

    def verify(self, values: ValueReference):
        if isinstance(values.values, dict):
            return self._verify_dict(values)
        elif not isinstance(values.values, list):
            if self.convert_str_to_singleton and isinstance(values.values, str):
                values.values = values.values = [values.values]
            elif values.values is None and values.action == OverrideActionEnum.Clear:
                pass
            else:
                msg = f"Type of {values.value_path} should be {list}, "
                msg += f"not {type(values.values)}"
                raise ConfigError(msg)
        elif not all(isinstance(el, str) for el in values.values):
            msg = f"Type of elements in {values.value_path} should be {str}"
            raise ConfigError(msg)
        path_actions = (OverrideActionEnum.AppendPath, OverrideActionEnum.PrependPath)
        if values.action in path_actions:
            msg = f"Option {values.value_path} of type {self.get_typename()} "
            msg += f"does not support operation {values.action.value}"
            raise ConfigError(msg)
        return ListOption.from_values(values, self.append_by_default)

    def finalize(self, values: ValueReference):
        val = values.values
        if isinstance(val, list):
            return val
        # We allow lists inheriting from strings
        if self.convert_str_to_singleton and isinstance(val, StringOption):
            val = val.finalize()
            return None if val is None else [val]
        # The following is the main case
        assert isinstance(val, ListOption)
        return val.finalize()
